#ifndef _STIM_UTIL_TOP_EXPORT_QUIRK_URL_H
#define _STIM_UTIL_TOP_EXPORT_QUIRK_URL_H

#include "stim/circuit/circuit.h"

namespace stim {

std::string export_quirk_url(const Circuit &circuit);

}  // namespace stim

#endif
